from typing import Literal

import mwapi  # type: ignore

from toolforge_i18n._user_agent import get_user_agent


def get_gender_by_user_name(user_name: str | None) -> Literal['m', 'f', 'n']:
    """Get the gender of a named user on Wikimedia sites.

    This gets the gender from Meta-Wiki –
    hopefully the user set it as a global preference,
    not just on one other wiki.
    ``None`` may be used to represent an unknown user (e.g. not logged in),
    who will be treated as having neuter gender.
    """
    if user_name is None:
        return 'n'
    session = mwapi.Session('https://meta.wikimedia.org', user_agent=get_user_agent())
    response = session.get(action='query', list=['users'], usprop=['gender'], ususers=[user_name], formatversion=2)
    mapping: dict[str, Literal['m', 'f', 'n']] = {
        'male': 'm',
        'female': 'f',
        'unknown': 'n',
    }
    return mapping[response['query']['users'][0]['gender']]



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html
import importlib.metadata

import sphinx_rtd_theme

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

metadata = importlib.metadata.metadata('toolforge_i18n')
project = metadata['name']
version = metadata['version']
author = metadata['author-email'].partition('<')[0].strip()

project_copyright = f'2024, {author} and contributors'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.intersphinx',
    'sphinx.ext.viewcode',
]

templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

intersphinx_mapping = {
    'flask': ('https://flask.palletsprojects.com/en/latest/', None),
    'markupsafe': ('https://markupsafe.palletsprojects.com/en/latest/', None),
    'toolforge': ('https://python-toolforge.readthedocs.io/en/latest/', None),
}


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

pygments_style = 'sphinx'
html_theme = 'sphinx_rtd_theme'
html_static_path = [sphinx_rtd_theme.get_html_theme_path()]



"""A collection of formatters for i18n messages.

You usually want to use I18nFormatter, which combines all the other
formatters in this module. Create it like this::

    formatter = I18nFormatter(locale_identifier='en',
                              get_gender=lambda arg: 'n')

And then use format strings such as::

    formatter.format('You have {count!p:one=one message:other={count} messages} from {users!l}.',
                     count=len(messages),
                     users=set(users))
    formatter.format('Reply to {user!g:m=him:f=her:n=them}?',
                     user=user)

See the documentation of the other formatter classes for details on
the supported conversions.

If the ``markupsafe`` module is available, then all the formatters defined
in this module become MarkupSafe-aware: if the format string is
``Markup``, then the returned formatted value will also be ``Markup``, and any
non-``Markup`` arguments will be escaped.
"""

import string
from collections.abc import Callable, Mapping, Sequence
from decimal import Decimal
from typing import Literal, TypeAlias, cast

import babel
import babel.lists

try:
    import markupsafe
except ModuleNotFoundError:
    markup_type = str

    class BaseI18nFormatter(string.Formatter):
        """Base class of our i18n formatters."""

        type_of_format_string: type[str] | None = str

        def __init__(self, **_kwargs: object):
            """Strip away all constructor arguments.

            ``string.Formatter`` complains if it gets any constructor
            arguments, but our individual constructors need to pass
            through their constructor arguments so the classes can be
            combined freely, so we throw away the arguments here.
            """
            super().__init__()

        def escape(self, s: str) -> str:
            """No-op function with a signature matching the escape below.

            Used by :py:class:`~HyperlinkFormatter` / :py:class:`~_Hyperlink`.
            """
            return s
else:
    markup_type = markupsafe.Markup  # type: ignore

    class BaseI18nFormatter(markupsafe.EscapeFormatter):  # type: ignore
        """Base class of our i18n formatters."""

        type_of_format_string: type[str] | None = None

        def __init__(self, **_kwargs: object):
            """Use a context-sensitive escape function.

            ``EscapeFormatter`` calls the escape function on all fields,
            since it expects to be reached through ``Markup.format()``,
            i.e. we know we’ll want to produce ``Markup`` and escape
            strings. We change this so that fields are only escaped if
            the format string was ``Markup`` in the first place.

            Also, throw away other constructor arguments, so that
            ``string.Formatter`` doesn’t complain about them.
            """

            def escape(s: object) -> markupsafe.Markup:
                assert self.type_of_format_string is not None
                if issubclass(self.type_of_format_string, markupsafe.Markup):
                    return markupsafe.Markup.escape(s)
                else:
                    # this cast is a lie but returning str seems to work in practice
                    return cast(markupsafe.Markup, s)

            super().__init__(escape=escape)

        def vformat(self, format_string: str, args: Sequence[object], kwargs: Mapping[str, object]) -> str:
            """Store the type of the format string and apply it to the result.

            ``string``.Formatter is partially implemented in C, so the
            format spec when formatting individual fields is always a
            string, never a ``Markup`` object. To still allow fields to
            properly format themselves as markup, store the type of
            the format string here. Also, apply it to the final result
            of string formatting.
            """

            self.type_of_format_string = type(format_string)
            try:
                ret = super().vformat(format_string, args, kwargs)
                return self.type_of_format_string(ret)
            finally:
                self.type_of_format_string = None


class PluralFormatter(BaseI18nFormatter):
    """A string formatter supporting a ``!p`` plural conversion.

    Format string examples::

        "I ate {count!p:0=no apples:one={count} apple:other={count} apples}."
        "{size} (0x{size:04X}) {size!p:one=bajt:two=bajtaj:few=bajty:other=bajtow}"

    For numeric values converted with ``!p``, the format spec is
    interpreted differently: it consists of a set of ``key=text`` specs,
    separated by colons. The key should be one of the CLDR plural rule
    tags, currently “zero”, “one”, “two”, “few”, “many”, or “other”,
    or an explicit value. The text for the matching value or tag,
    according to the plural rules of the locale specified in the
    constructor, is substituted into the message. Attempting to
    convert non-numeric values with ``!p`` is an error.

    Note that most languages do not use all possible tags, and only
    exactly those tags used in a language should occur in the format
    string. For example, even though there is a “zero” tag, English
    only uses the “one” and “other” ones, and to make a special case
    for a value of zero with a ``PluralFormatter('en')``, you need to use
    the key “0”, not “zero”. On the other hand, failing to specify all
    tags used in a language may make the formatter raise a KeyError:
    for instance, if the first example above used the key “1” instead
    of “one”, then it would fail when given a count of ``-1`` or ``1.0``.

    Value keys always take precedence over tag keys, no matter in
    which order they are specified in the format spec. To match the
    value, they must be identical to the ``str()`` of the value: for
    instance, a “1” key will not match a ``1.0`` value or vice versa.
    """

    def __init__(self, *, locale_identifier: str, **kwargs: object):
        """The locale identifier must be understood by ``Locale.parse``."""
        self.locale = babel.Locale.parse(locale_identifier)
        super().__init__(locale_identifier=locale_identifier, **kwargs)

    def convert_field(self, value: float | Decimal, conversion: str | None) -> object:
        if conversion == 'p':
            assert self.type_of_format_string is not None
            return _Plural(value, self.locale, self.type_of_format_string)
        return super().convert_field(value, conversion)


class _Plural:
    """Wrapper around a numeric value with special formatting.

    This class formats itself as described in :py:class:`~PluralFormatter`.
    """

    def __init__(self, value: float | Decimal, locale: babel.Locale, type_of_format_spec: type[str]):
        self.value = value
        self.locale = locale
        self.type_of_format_spec = type_of_format_spec

    def __format__(self, format_spec: str) -> str:
        format_spec = self.type_of_format_spec(format_spec)
        plurals = format_spec.split(':')
        value_eq = str(self.value) + '='
        for plural in plurals:
            if plural.startswith(value_eq):
                return plural[len(value_eq) :]
        tag = self.locale.plural_form(self.value)
        tag_eq = tag + '='
        for plural in plurals:
            if plural.startswith(tag_eq):
                return plural[len(tag_eq) :]
        # fall back to "other"
        for plural in plurals:
            if plural.startswith('other='):
                return plural[len('other=') :]
        raise KeyError(f'No plurals for tag "{tag}" or "other" found in format spec "{format_spec}"!')


class CommaSeparatedListFormatter(BaseI18nFormatter):
    """A string formatter supporting a ``!l`` list conversion.

    Format string example::

        "We went to {cities!l}."

    For iterable values converted with ``!l``, the format spec is applied
    to each list element. Afterwards, the list elements are joined
    into a standard list using the locale specified in the
    constructor. (For English, this means separating most items with
    an ASCII comma plus a space, and the final two with an extra
    “and”; Chinese and Japanese, for instance, use a fullwidth comma
    instead.) Attempting to convert non-iterable values with ``!l`` is an
    error.
    """

    def __init__(self, *, locale_identifier: str, **kwargs: object):
        """The locale identifier must be understood by ``Locale.parse``."""
        self.locale = babel.Locale.parse(locale_identifier)
        # convert Babel’s list patterns to Markup
        for style in self.locale.list_patterns:
            for key, pattern in self.locale.list_patterns[style].items():
                self.locale.list_patterns[style][key] = markup_type(pattern)
        super().__init__(locale_identifier=locale_identifier, **kwargs)

    def convert_field(self, value: Sequence[object], conversion: str | None) -> object:
        if conversion == 'l':
            assert self.type_of_format_string is not None
            return _CommaSeparatedList(value, self.locale, self.type_of_format_string, self.format_field)
        return super().convert_field(value, conversion)


class _CommaSeparatedList:
    """Wrapper around a list with special formatting.

    This class formats itself as described in :py:class:`~CommaSeparatedListFormatter`.
    """

    def __init__(
        self,
        value: Sequence[object],
        locale: babel.Locale,
        type_of_format_spec: type[str],
        format_function: Callable[[object, str], object],
    ):
        self.value = value
        self.locale = locale
        self.type_of_format_spec = type_of_format_spec
        self.format_function = format_function

    def __format__(self, format_spec: str) -> str:
        format_spec = self.type_of_format_spec(format_spec)
        formatted_values = [markup_type(self.format_function(item, format_spec)) for item in self.value]
        return babel.lists.format_list(formatted_values, locale=self.locale)


GetGender: TypeAlias = Callable[[object], Literal['m', 'f', 'n']]


class GenderFormatter(BaseI18nFormatter):
    """A string formatter supporting a ``!g`` grammatical gender conversion.

    Format string examples::

        "Leave a message on {user!g:m=his:f=her:n=their} talk page."
        "Ci dispiace, ma non sei {user!g:m=autorizzato:f=autorizzata:n=autorizzato/a} a usare il caricamento di massa."

    The formatted value, which can be anything as far as this
    formatter is concerned, is passed into a function specified in the
    constructor, which should return one of the values ``"m"``, ``"f"``, or
    ``"n"``, to select the grammatically masculine, feminine, or neutral
    replacement, respectively. The format spec specifies these three
    replacements separated by colons. Gender values not specified in
    the format spec fall back to ``"m"``.
    """

    def __init__(self, *, get_gender: GetGender, **kwargs: object):
        self.get_gender = get_gender
        super().__init__(get_gender=get_gender, **kwargs)

    def convert_field(self, value: object, conversion: str | None) -> object:
        if conversion == 'g':
            assert self.type_of_format_string is not None
            return _Gender(value, self.get_gender, self.type_of_format_string)
        return super().convert_field(value, conversion)


class _Gender:
    """Wrapper around a value with special formatting.

    This class formats itself as described in :py:class:`~GenderFormatter`.
    """

    def __init__(self, value: object, get_gender: GetGender, type_of_format_spec: type[str]):
        self.value = value
        self.get_gender = get_gender
        self.type_of_format_spec = type_of_format_spec

    def __format__(self, format_spec: str) -> str:
        format_spec = self.type_of_format_spec(format_spec)
        replacements = format_spec.split(':')
        if len(replacements) == 1:
            replacement = replacements[0]
            assert replacement.startswith('m=')
            return replacement[len('m=') :]
        gender = self.get_gender(self.value)
        gender_eq = gender + '='
        for replacement in replacements:
            if replacement.startswith(gender_eq):
                return replacement[len(gender_eq) :]
        # fall back to "m"
        for replacement in replacements:
            if replacement.startswith('m='):
                return replacement[len('m=') :]
        raise KeyError(f'No replacement for gender "{gender}" or "m" found in format spec "{format_spec}"!')


class HyperlinkFormatter(BaseI18nFormatter):
    """A string formatter supporting an ``!h`` hyperlink conversion.

    Format string example::

        "You need to {url!h:log in} before you can edit."

    The formatted value is interpreted as the ``href`` attribute of an
    HTML ``<a>`` element, whose inner HTML is given by the format spec.
    """

    def convert_field(self, value: str, conversion: str | None) -> object:
        if conversion == 'h':
            assert self.type_of_format_string is not None
            return _Hyperlink(value, self.type_of_format_string, self.escape)
        return super().convert_field(value, conversion)


class _Hyperlink:
    """Wrapper around a URL with special formatting.

    This class formats itself as described in :py:class`~HyperlinkFormatter`.
    """

    def __init__(self, value: str, type_of_format_spec: type[str], escape: Callable[[str], str]):
        self.value = value
        self.type_of_format_spec = type_of_format_spec
        self.escape = escape

    def __format__(self, format_spec: str) -> str:
        format_spec = self.type_of_format_spec(format_spec)
        # turn the value into the type of the format spec,
        # so that if format spec is str and value is Markup,
        # the value doesn’t escape everything around it –
        # but first escape the value, in case it’s str
        # and the format spec is Markup
        value = self.type_of_format_spec(self.escape(self.value))
        return (
            self.type_of_format_spec(r'<a href="')
            + value
            + self.type_of_format_spec(r'">')
            + format_spec
            + self.type_of_format_spec(r'</a>')
        )


class I18nFormatter(PluralFormatter, CommaSeparatedListFormatter, GenderFormatter, HyperlinkFormatter):  # type: ignore[misc]
    """A string formatter supporting ``!p`` (plural), ``!l`` (list),
    ``!g`` (gender) and ``!h`` (hyperlink) conversions.

    See :py:class:`~PluralFormatter`, :py:class:`~CommaSeparatedListFormatter`,
    :py:class:`~HyperlinkFormatter` and :py:class:`~GenderFormatter` for details.
    """



from contextlib import suppress as _suppress

with _suppress(ModuleNotFoundError):
    from pytest import register_assert_rewrite as _register_assert_rewrite  # noqa: PT013

    _register_assert_rewrite('toolforge_i18n._translations_checks')

from toolforge_i18n._formatters import (
    CommaSeparatedListFormatter,
    GenderFormatter,
    HyperlinkFormatter,
    I18nFormatter,
    PluralFormatter,
)
from toolforge_i18n._get_gender import get_gender_by_user_name
from toolforge_i18n._language_info import lang_autonym, lang_bcp47_to_mw, lang_dir, lang_fallbacks, lang_mw_to_bcp47
from toolforge_i18n._translations import TranslationsConfig, language_code_to_babel, load_translations
from toolforge_i18n._user_agent import get_user_agent, set_user_agent

with _suppress(ModuleNotFoundError):
    from toolforge_i18n._flask import (
        ToolforgeI18n,
        add_lang_if_needed,
        interface_language_code_from_request,
        message,
        pop_html_lang,
        push_html_lang,
    )



from collections.abc import Callable
from typing import cast

import flask
import werkzeug
from markupsafe import Markup

from toolforge_i18n._formatters import I18nFormatter
from toolforge_i18n._language_info import lang_bcp47_to_mw, lang_dir, lang_fallbacks, lang_mw_to_bcp47
from toolforge_i18n._translations import load_translations


def init_html_language_codes() -> None:
    """Initialize the stack of HTML language codes.

    The stack is used to track whether a message needs to be wrapped
    in a new HTML element to set the lang= and dir= attributes,
    or whether it’s in the same language as the surrounding markup.
    It’s maintained by :py:func:`~message`, :py:func:`~push_html_lang` and :py:func:`~pop_html_lang`.
    """
    flask.g.html_language_codes = []


def push_html_lang(language_code: str) -> Markup:
    """Push an HTML language code to the stack.

    Many tools will not need to call this,
    as it’s called by the :py:func:`~message` function automatically.
    However, if you also add localized text from other sources than messages,
    you should call this function with the MediaWiki language code you are using;
    for example, in a Jinja2 template::

        <span {{ push_html_lang(label.language) }}>
          {{ label.value }}
        </span{{ pop_html_lang(label.language) }}>
    """
    html_language_code = lang_mw_to_bcp47(language_code)
    flask.g.html_language_codes.append(html_language_code)
    return Markup(r'lang="{}" dir="{}"').format(html_language_code, lang_dir(html_language_code))


def add_lang_if_needed(message: Markup, language_code: str) -> Markup:
    """Wrap the given message in a language-tagged element if necessary.

    Given a (formatted) message in a certain language (MediaWiki language code),
    wrap it in a ``<span>`` with ``lang=`` and ``dir=`` attributes
    if the current language on top of the stack is different.
    Note that :py:func:`~message` calls this function automatically,
    so you generally don’t need to use this function yourself.
    """
    if flask.g.html_language_codes and flask.g.html_language_codes[-1] == language_code:
        return message
    return Markup('<span {}>{}</span{}>').format(push_html_lang(language_code), message, pop_html_lang(language_code))


def pop_html_lang(language_code: str) -> Markup:
    """Pop an HTML language code from the stack.

    See :py:func:`~push_html_lang` for details.
    """
    html_language_code = lang_mw_to_bcp47(language_code)
    assert flask.g.html_language_codes.pop() == html_language_code
    return Markup(r'')


def assert_html_language_codes_empty(response: werkzeug.Response) -> werkzeug.Response:
    """Assert that the stack of HTML language codes is depleted.

    This is called at the end of a request;
    if the assertion fails, some :py:func:`~push_html_lang` call is missing
    a corresponding :py:func:`~pop_html_lang` call.
    """
    assert flask.g.html_language_codes == []
    return response


def interface_language_code_from_request(translations: dict[str, dict[str, str]]) -> str:
    """Default implementation to determine the language code of a request.

    This function supports the ``?uselang=`` URL parameter
    and otherwise determines the language based on the request’s ``Accept-Language`` header.
    You may want to override this method to implement a persistent language preference;
    to keep the features mentioned above,
    your implementation should generally look like this::

        from toolforge_i18n import interface_language_code_from_request

        def interface_language_code(translations):
            # ?uselang= takes precedence if present
            if 'uselang' in flask.request.args:
                return interface_language_code_from_request(translations)
            # try persistent language preference (e.g. from flask.session) next
            # ...
            # finally, fall back to Accept-Language:
            return interface_language_code_from_request(translations)

        # ...later, pass the implementation into ToolforgeI18n:
        i18n = ToolforgeI18n(app, interface_language_code)
    """
    if 'uselang' in flask.request.args:
        return flask.request.args['uselang']
    available_bcp47_languages = [lang_mw_to_bcp47(code) for code in translations]
    best_bcp47_language = flask.request.accept_languages.best_match(available_bcp47_languages, 'en')
    return lang_bcp47_to_mw(best_bcp47_language)


def _message_with_language(message_code: str) -> tuple[Markup, str]:
    interface_language_code = cast(str, flask.g.interface_language_code)
    language_codes = [interface_language_code, *lang_fallbacks(interface_language_code), 'en']
    translations = flask.current_app.extensions['toolforge_i18n'].translations
    for language_code in language_codes:
        try:
            text = translations[language_code][message_code]
        except LookupError:
            continue
        else:
            return Markup(text), language_code
    raise ValueError(f'Message {message_code} not found in {language_codes}')


def _message_qqx(message_code: str, **kwargs: object) -> Markup:
    message = Markup('(')
    message += message_code
    if kwargs:
        message += ': '
        first = True
        for key, value in kwargs.items():
            if first:
                first = False
            else:
                message += ', '
            message += key
            message += '='
            message += repr(value)
    message += ')'
    return message


def message(message_code: str, **kwargs: object) -> Markup:
    """Format an interface message in the user interface language.

    The kwargs may contain (named) arguments,
    using the argument names defined in :py:attr:`~TranslationsConfig.variables`.

    This method is available as a template global, and is usually used there
    (but may also be imported and called from Python code).
    """
    if flask.g.qqx:
        return _message_qqx(message_code, **kwargs)
    message, language = _message_with_language(message_code)
    if kwargs:
        config = flask.current_app.extensions['toolforge_i18n'].config
        formatter = I18nFormatter(
            locale_identifier=config.language_code_to_babel(language),
            get_gender=config.get_gender,
        )
        # I18nFormatter returns Markup given Markup
        message = cast(Markup, formatter.format(message, **kwargs))
    return add_lang_if_needed(message, language)


class ToolforgeI18n:
    """Flask extension for toolforge_i18n.

    Basic usage::

        app = flask.Flask(__name__)
        i18n = ToolforgeI18n(app)
    """

    def __init__(
        self,
        app: flask.Flask | None = None,
        interface_language_code: Callable[[dict[str, dict[str, str]]], str] = interface_language_code_from_request,
    ):
        import tool_translations_config

        self.config = tool_translations_config.config
        self.translations, self.documentation = load_translations(self.config)
        self.interface_language_code = interface_language_code
        if app is not None:
            self.init_app(app)

    def init_app(self, app: flask.Flask) -> None:
        app.extensions['toolforge_i18n'] = self
        app.add_template_global(message)
        app.add_template_filter(lang_mw_to_bcp47)
        app.add_template_global(push_html_lang)
        app.add_template_global(pop_html_lang)
        app.before_request(init_html_language_codes)
        app.after_request(assert_html_language_codes_empty)

        @app.before_request
        def init_interface_language_code() -> None:
            interface_language_code = self.interface_language_code(self.translations)
            if interface_language_code == 'qqx':
                flask.g.interface_language_code = 'en'
                flask.g.qqx = True
            else:
                flask.g.interface_language_code = interface_language_code
                flask.g.qqx = False



from typing import Literal, Never

import pytest
from markupsafe import Markup

from toolforge_i18n._formatters import (
    CommaSeparatedListFormatter,
    GenderFormatter,
    HyperlinkFormatter,
    I18nFormatter,
    PluralFormatter,
)


@pytest.mark.parametrize(
    'count, expected',
    [
        (1, 'I ate 1 apple.'),
        (2, 'I ate 2 apples.'),
        (0, 'I ate no apples.'),
    ],
)
def test_PluralFormatter_en(count: int, expected: str) -> None:
    plural_formatter = PluralFormatter(locale_identifier='en')
    assert (
        plural_formatter.format('I ate {count!p:0=no apples:one={count} apple:other={count} apples}.', count=count)
        == expected
    )


@pytest.mark.parametrize(
    'size, expected',
    [
        (1, '1 (0x0001) bajt'),
        (2, '2 (0x0002) bajtaj'),
        (3, '3 (0x0003) bajty'),
        (4, '4 (0x0004) bajty'),
        (5, '5 (0x0005) bajtow'),
    ],
)
def test_PluralFormatter_hsb(size: int, expected: str) -> None:
    plural_formatter = PluralFormatter(locale_identifier='hsb')
    assert (
        plural_formatter.format('{size} (0x{size:04X}) {size!p:one=bajt:two=bajtaj:few=bajty:other=bajtow}', size=size)
        == expected
    )


@pytest.mark.parametrize(
    'num, expected',
    [
        (0, 'zero'),
        (1, 'one'),
        (2, 'two'),
        (3, 'other'),
        (0.0, 'other'),
    ],
)
def test_PluralFormatter_explicit(num: float, expected: str) -> None:
    plural_formatter = PluralFormatter(locale_identifier='en')
    assert plural_formatter.format('{num!p:0=zero:1=one:2=two:other=other}', num=num) == expected


def test_PluralFormatter_explicit_takes_precedence() -> None:
    plural_formatter = PluralFormatter(locale_identifier='en')
    assert plural_formatter.format('{num!p:other=other:0=zero}', num=0) == 'zero'


@pytest.mark.parametrize(
    'format_spec, exc_type',
    [
        ('{!p}', KeyError),
        ('{!p:}', KeyError),
        ('{!p:0=missing plural for other}', KeyError),
    ],
)
def test_PluralFormatter_invalid_format_spec(format_spec: str, exc_type: type[BaseException]) -> None:
    plural_formatter = PluralFormatter(locale_identifier='en')
    with pytest.raises(exc_type):
        plural_formatter.format(format_spec, 1)


def test_PluralFormatter_fallback() -> None:
    plural_formatter = PluralFormatter(locale_identifier='en')
    formatted = plural_formatter.format('prefix {val!p:0=zero:other="{val}"} suffix', val=1)
    assert formatted == 'prefix "1" suffix'


@pytest.mark.parametrize(
    'format_type, adjective_type, expected_str',
    [
        (str, str, 'A single "<strong>cool</strong>" item.'),
        (Markup, Markup, 'A single "<strong>cool</strong>" item.'),
        (str, Markup, 'A single "<strong>cool</strong>" item.'),
        (Markup, str, 'A single "&lt;strong&gt;cool&lt;/strong&gt;" item.'),
    ],
)
def test_PluralFormatter_MarkupSafe(format_type: type[str], adjective_type: type[str], expected_str: str) -> None:
    plural_formatter = PluralFormatter(locale_identifier='en')
    formatted = plural_formatter.format(
        format_type('{count!p:one=A single "{adjective}" item.:other=Several "{adjective}" items.}'),
        count=1,
        adjective=adjective_type('<strong>cool</strong>'),
    )
    assert formatted == expected_str
    assert type(formatted) is format_type


@pytest.mark.parametrize(
    'codecs, expected',
    [
        (['FLAC'], 'FLAC.'),
        (['FLAC', 'OGG'], 'FLAC and OGG.'),
        (['FLAC', 'OGG', 'OPUS'], 'FLAC, OGG, and OPUS.'),
    ],
)
def test_CommaSeparatedListFormatter_en(codecs: list[str], expected: str) -> None:
    comma_separated_list_formatter = CommaSeparatedListFormatter(locale_identifier='en')
    assert comma_separated_list_formatter.format('{codecs!l}.', codecs=codecs) == expected


@pytest.mark.parametrize(
    'codecs, expected',
    [
        (['FLAC'], 'FLAC.'),
        (['FLAC', 'OGG'], 'FLAC和OGG.'),
        (['FLAC', 'OGG', 'OPUS'], 'FLAC、OGG和OPUS.'),
    ],
)
def test_CommaSeparatedListFormatter_zh(codecs: list[str], expected: str) -> None:
    comma_separated_list_formatter = CommaSeparatedListFormatter(locale_identifier='zh')
    assert comma_separated_list_formatter.format('{codecs!l}.', codecs=codecs) == expected


def test_CommaSeparatedListFormatter_formats_list_items_with_format_spec() -> None:
    comma_separated_list_formatter = CommaSeparatedListFormatter(locale_identifier='en')
    assert (
        comma_separated_list_formatter.format(
            'The binaries are {sizes!l:04d} bytes large.', sizes=[64, 128, 256, 1024, 4096]
        )
        == 'The binaries are 0064, 0128, 0256, 1024, and 4096 bytes large.'
    )


@pytest.mark.parametrize(
    'format_type, item_type, expected_str',
    [
        (str, str, 'I like <i>Python</i>.'),
        (Markup, Markup, 'I like <i>Python</i>.'),
        (str, Markup, 'I like <i>Python</i>.'),
        (Markup, str, 'I like &lt;i&gt;Python&lt;/i&gt;.'),
    ],
)
def test_CommaSeparatedListFormatter_MarkupSafe_one_item(
    format_type: type[str], item_type: type[str], expected_str: str
) -> None:
    comma_separated_list_formatter = CommaSeparatedListFormatter(locale_identifier='en')
    formatted = comma_separated_list_formatter.format(
        format_type('I like {list!l}.'),
        list=[
            item_type('<i>Python</i>'),
        ],
    )
    assert formatted == expected_str
    assert type(formatted) is format_type


@pytest.mark.parametrize(
    'format_type, item_one_type, item_two_type, expected_str',
    [
        (str, str, str, 'I like <i>Python</i> and <i>Wikidata</i>.'),
        (Markup, Markup, Markup, 'I like <i>Python</i> and <i>Wikidata</i>.'),
        (str, Markup, Markup, 'I like <i>Python</i> and <i>Wikidata</i>.'),
        (Markup, str, str, 'I like &lt;i&gt;Python&lt;/i&gt; and &lt;i&gt;Wikidata&lt;/i&gt;.'),
        (Markup, Markup, str, 'I like <i>Python</i> and &lt;i&gt;Wikidata&lt;/i&gt;.'),
        (Markup, str, Markup, 'I like &lt;i&gt;Python&lt;/i&gt; and <i>Wikidata</i>.'),
    ],
)
def test_CommaSeparatedListFormatter_MarkupSafe_two_items(
    format_type: type[str], item_one_type: type[str], item_two_type: type[str], expected_str: str
) -> None:
    comma_separated_list_formatter = CommaSeparatedListFormatter(locale_identifier='en')
    formatted = comma_separated_list_formatter.format(
        format_type('I like {list!l}.'),
        list=[
            item_one_type('<i>Python</i>'),
            item_two_type('<i>Wikidata</i>'),
        ],
    )
    assert formatted == expected_str
    assert type(formatted) is format_type


@pytest.mark.parametrize(
    'format_type, item_one_type, item_two_type, item_three_type, expected_str',
    [
        (str, str, str, str, 'I like <i>Python</i>, <i>Wikidata</i>, and <i>Music</i>.'),
        (Markup, Markup, Markup, Markup, 'I like <i>Python</i>, <i>Wikidata</i>, and <i>Music</i>.'),
        (Markup, str, Markup, str, 'I like &lt;i&gt;Python&lt;/i&gt;, <i>Wikidata</i>, and &lt;i&gt;Music&lt;/i&gt;.'),
    ],
)
def test_CommaSeparatedListFormatter_MarkupSafe_three_items(
    format_type: type[str],
    item_one_type: type[str],
    item_two_type: type[str],
    item_three_type: type[str],
    expected_str: str,
) -> None:
    comma_separated_list_formatter = CommaSeparatedListFormatter(locale_identifier='en')
    formatted = comma_separated_list_formatter.format(
        format_type('I like {list!l}.'),
        list=[
            item_one_type('<i>Python</i>'),
            item_two_type('<i>Wikidata</i>'),
            item_three_type('<i>Music</i>'),
        ],
    )
    assert formatted == expected_str
    assert type(formatted) is format_type


@pytest.mark.parametrize(
    'gender, expected',
    [
        ('m', 'Thank him?'),
        ('f', 'Thank her?'),
        ('n', 'Thank them?'),
    ],
)
def test_GenderFormatter(gender: Literal['m', 'f', 'n'], expected: str) -> None:
    user = 'opaque value'

    def get_gender(value: object) -> Literal['m', 'f', 'n']:
        assert value == user
        return gender

    gender_formatter = GenderFormatter(get_gender=get_gender)
    assert gender_formatter.format('Thank {user!g:m=him:f=her:n=them}?', user=user) == expected


@pytest.mark.parametrize(
    'gender',
    [
        ('m'),
        ('f'),
        ('n'),
    ],
)
def test_GenderFormatter_fallback(gender: Literal['m', 'f', 'n']) -> None:
    user = 'opaque value'

    def get_gender(value: object) -> Literal['m', 'f', 'n']:
        assert value == user
        return gender

    gender_formatter = GenderFormatter(get_gender=get_gender)
    assert (
        gender_formatter.format(
            'Thank {user!g:m=you:o=unused}!',  # o(ther) only included in format spec to make this test different from _unused below
            user=user,
        )
        == 'Thank you!'
    )


def test_GenderFormatter_unused() -> None:
    user = 'opaque value'

    def get_gender(_value: object) -> Never:
        # pointless to call get_gender() when format spec only has one replacement
        raise AssertionError('Should not be called at all!')

    gender_formatter = GenderFormatter(get_gender=get_gender)
    assert gender_formatter.format('Thank {user!g:m=you}!', user=user) == 'Thank you!'


@pytest.mark.parametrize(
    'format_type, adjective_type, expected_str',
    [
        (str, str, 'Please respect <em>their</em> <strong>correct</strong> pronouns.'),
        (Markup, Markup, 'Please respect <em>their</em> <strong>correct</strong> pronouns.'),
        (str, Markup, 'Please respect <em>their</em> <strong>correct</strong> pronouns.'),
        (Markup, str, 'Please respect <em>their</em> &lt;strong&gt;correct&lt;/strong&gt; pronouns.'),
    ],
)
def test_GenderFormatter_MarkupSafe(format_type: type[str], adjective_type: type[str], expected_str: str) -> None:
    gender_formatter = GenderFormatter(get_gender=lambda _: 'n')
    formatted = gender_formatter.format(
        format_type(
            'Please respect {user!g:m=<em>his</em> {adjective}:f=<em>her</em> {adjective}:n=<em>their</em> {adjective}} pronouns.'
        ),
        user='user',
        adjective=adjective_type('<strong>correct</strong>'),
    )
    assert formatted == expected_str
    assert type(formatted) is format_type


def test_HyperlinkFormatter() -> None:
    hyperlink_formatter = HyperlinkFormatter()
    assert (
        hyperlink_formatter.format(
            'You need to {url!h:log in} before you can edit.',
            url='/login',
        )
        == 'You need to <a href="/login">log in</a> before you can edit.'
    )


@pytest.mark.parametrize(
    'format_type, url_type, expected_str',
    [
        (str, str, 'You need to <a href="/login"">log <em>in</em></a> before you can edit.'),
        (Markup, Markup, 'You need to <a href="/login"">log <em>in</em></a> before you can edit.'),
        (str, Markup, 'You need to <a href="/login"">log <em>in</em></a> before you can edit.'),
        (Markup, str, 'You need to <a href="/login&#34;">log <em>in</em></a> before you can edit.'),
    ],
)
def test_HyperlinkFormatter_MarkupSafe(format_type: type[str], url_type: type[str], expected_str: str) -> None:
    hyperlink_formatter = HyperlinkFormatter()
    formatted = hyperlink_formatter.format(
        format_type('You need to {url!h:log <em>in</em>} before you can edit.'),
        url=url_type('/login"'),
    )
    assert formatted == expected_str
    assert type(formatted) is format_type


flac = '<abbr title="Free Lossless Audio Codec">FLAC</abbr>'


@pytest.mark.parametrize(
    'formats, user, expected',
    [
        ([Markup(flac)], 'Keith', f'His preferred <a href="/wiki/Format">format</a> is {flac}.'),
        (
            [Markup(flac), '"OGG"'],
            'Keira',
            f'Her preferred <a href="/wiki/Format">formats</a> are {flac} and &#34;OGG&#34;.',
        ),
        (
            [Markup(flac), '"OGG"', 'OPUS'],
            'Kim',
            f'Their preferred <a href="/wiki/Format">formats</a> are {flac}, &#34;OGG&#34;, and OPUS.',
        ),
    ],
)
def test_I18nFormatter_en(formats: list[str], user: str, expected: str) -> None:
    genders = {
        'Keith': 'm',
        'Keira': 'f',
        'Kim': 'n',
    }
    i18n_formatter = I18nFormatter(locale_identifier='en', get_gender=lambda value: genders[value])
    formatted = i18n_formatter.format(
        Markup(
            '{user!g:m=His:f=Her:n=Their} preferred {count!p:one={url!h:format} is:other={url!h:formats} are} {formats!l}.'
        ),
        user=user,
        count=len(formats),
        formats=formats,
        url='/wiki/Format',
    )
    assert formatted == expected
    assert isinstance(formatted, Markup)



from typing import Literal, TypedDict, cast

import mwapi  # type: ignore

from toolforge_i18n._user_agent import get_user_agent

# for easier typing below, pretend that _LanguageInfo members are optional,
# and that _language_info and _by_bcp47 are always dicts
# (they are lazily initialized in _load_language_info)


class _LanguageInfo(TypedDict, total=False):
    bcp47: str
    dir: Literal['ltr', 'rtl']
    autonym: str
    fallbacks: list[str]


_language_info: dict[str, _LanguageInfo] = cast(dict[str, _LanguageInfo], None)
_by_bcp47: dict[str, str] = cast(dict[str, str], None)


def _load_language_info() -> None:
    global _language_info, _by_bcp47
    if _language_info is not None:
        return

    session = mwapi.Session(
        'https://meta.wikimedia.org',
        user_agent=get_user_agent(),
    )
    _language_info = {}
    for response in session.get(
        continuation=True,
        action='query',
        meta='languageinfo',
        liprop=['autonym', 'bcp47', 'dir', 'fallbacks'],
        formatversion='2',
    ):
        _language_info.update(response['query']['languageinfo'])

    _by_bcp47 = {}
    for code, language in _language_info.items():
        _by_bcp47[language['bcp47']] = code


# lang_mw_to_bcp47 and lang_bcp47_to_mw have both “source” and “destination” type in their name,
# because we need both directions and the names would otherwise be confusing;
# the other functions (lang_autonym, lang_dir, lang_fallbacks)
# all take MediaWiki language codes as the “source”


def lang_mw_to_bcp47(code: str) -> str:
    """Get the BCP-47 language code of the given MediaWiki language code."""
    _load_language_info()
    return _language_info.get(code, {}).get('bcp47', code)


def lang_bcp47_to_mw(code: str) -> str:
    """Get the MediaWiki language code of the given BCP-47 language code."""
    _load_language_info()
    return _by_bcp47.get(code, code)


def lang_autonym(code: str) -> str | None:
    """Get the autonym of the given language code, according to MediaWiki."""
    _load_language_info()
    return _language_info.get(code, {}).get('autonym')


def lang_dir(code: str) -> Literal['ltr', 'rtl', 'auto']:
    """Get the directionality of the given language code, according to MediaWiki."""
    _load_language_info()
    return _language_info.get(code, {}).get('dir', 'auto')


def lang_fallbacks(code: str) -> list[str]:
    """Get the fallback languages of the given language code, according to MediaWiki."""
    _load_language_info()
    return _language_info.get(code, {}).get('fallbacks', [])



import requests

_user_agent = None


def get_user_agent() -> str:
    """Get the user agent string used by toolforge_i18n.

    The user agent string may be set by :py:func:`~set_user_agent`;
    otherwise, try to get a user agent previously set up
    by :py:func:`toolforge.set_user_agent`.

    Code outside of toolforge_i18n generally shouldn’t use this function.
    """
    if _user_agent is not None:
        return _user_agent
    toolforge_user_agent = requests.utils.default_user_agent()
    if 'toolforge' in toolforge_user_agent:
        return toolforge_user_agent
    raise RuntimeError(
        'Could not determine user agent, '
        'call either toolforge.set_user_agent() '
        'or toolforge_i18n.set_user_agent()'
    )


def set_user_agent(user_agent: str) -> None:
    """Set the user agent string used by toolforge_i18n.

    Most tools should call :py:func:`toolforge.set_user_agent` instead,
    which also sets the user agent for other code.
    It is typically called during early initialization.

    See the `User-Agent policy <https://w.wiki/9mMA>`_ for the format.
    """
    global _user_agent
    _user_agent = user_agent



import json
import os
import re
from collections.abc import Callable, Mapping, Sequence
from dataclasses import dataclass, field
from typing import Any, Literal

import babel

from toolforge_i18n._get_gender import get_gender_by_user_name


def language_code_to_babel(code: str) -> str:
    """Default implementation to map a MediaWiki language code to Babel.

    This implementation is conservative and only maps language codes
    where Babel has an alternative that does not lose any information
    (at least as far as toolforge_i18n is concerned).
    MediaWiki also supports many language codes that (as far as I know)
    have no lossless equivalent in Babel,
    such as (at the time of writing) sh-latn (Serbo-Croatian in Latin script).
    If your tool is translated into one of those languages,
    you will have to configure a custom ``language_code_to_babel`` implementation
    in your ``tool_translations_config`` and pick some lossy fallback
    (e.g. hr, Croatian, for sh-latn).
    Your implementation should generally delegate to this one first, for instance::

        def language_code_to_babel(code: str) -> str:
            mapped = toolforge_i18n.language_code_to_babel(code)
            if mapped != code:
                return mapped
            return {
                'sh-latn': 'hr',
                # ...
            }.get(code, code.partition('-')[0])
    """

    return {
        'be-tarask': 'be_TARASK',
        'de-formal': 'de',  # the formal/informal distinction is irrelevant for I18nFormatter
        'nds-nl': 'nds_NL',
        'kk-cyrl': 'kk',  # Cyrl implied: see e.g. babel.Locale('kk').languages['en']
        'ko-kp': 'ko_KP',
        'ks-arab': 'ks_Arab',
        'ku-latn': 'ku',  # Latn implied: see e.g. babel.Locale('ku').languages['en']
        'ms-arab': 'ms_Arab',
        'nl-informal': 'nl',  # the formal/informal distinction is irrelevant for I18nFormatter
        'pt-br': 'pt_BR',
        'sr-ec': 'sr_Cyrl',
        'sr-el': 'sr_Latn',
        'tg-cyrl': 'tg',  # Cyrl implied: see e.g. babel.Locale('tg').languages['en']
        'tt-cyrl': 'tt',  # Cyrl implied: see e.g. babel.Locale('tt').languages['en']
        'ug-arab': 'ug',  # Arab implied: see e.g. babel.Locale('ug').languages['en']
        'yue-hant': 'yue_Hant',
        'zh-hans': 'zh_Hans',
        'zh-hant': 'zh_Hant',
    }.get(code, code)


@dataclass
class TranslationsConfig:
    """Configuration for loading message translations.

    To use this library, a tool should define a ``tool_translations_config`` module
    which exports a ``config`` member of this type, like so::

        # tool_translations_config.py
        import TranslationsConfig from toolforge_i18n
        config = TranslationsConfig(
            # ...
        )

    The most important config to define is :py:attr:`~variables`,
    which most tools will need (unless all your messages have no variables);
    the others may or may not be necessary depending on the tool.
    """

    directory: str = 'i18n/'
    """The path to the directory to load message files from."""

    variables: Mapping[str, Sequence[str]] = field(default_factory=dict)
    """Variable names used in messages.

    The source messages use $1, $2 etc.,
    but the Python format strings use named variables,
    whose names are specified here.
    The variable name (or its prefix) encodes the type:

    * ``url``, ``url_*`` - hyperlink: ``[$1 text]`` => ``{url!h:text}``
    * ``user_name``, ``user_name_*`` - gender: ``{{GENDER:$1|he|she|they}}`` => ``{user_name!g:m=he:f=she:n=they}``
    * ``num``, ``num_*`` - plural: ``{{PLURAL:$1|one egg|$1 eggs}}`` => ``{num_eggs!p:one=one egg:other={num_eggs} eggs}``
    * ``list``, ``list_*`` - list: ``$1`` => ``{list_chicken_names!l}``
    * anything else - markup without further formatting: ``$1`` => ``{description}``
    """

    derived_messages: Mapping[str, tuple[str, Callable[[str], str]]] = field(default_factory=dict)
    """Messages that are derived from other messages.

    The key is a message key that is not expected in the JSON files,
    but that is instead generated by taking another message
    (whose key is the first element of the tuple)
    and sending it through the callable in the second element of the tuple.
    Examples for that callable include the identity function
    (to copy a message) or simple case transformations.
    """

    language_code_to_babel: Callable[[str], str] = language_code_to_babel
    """Mapping from MediaWiki to Babel language codes.

    Message files use MediaWiki language codes,
    which are not always standard language codes;
    additionally, MediaWiki supports many languages that Babel does not,
    even when the language code is standard.
    You will need to map these codes to a supported alternative
    if you have any translations for such language codes.
    """

    allowed_html_elements: dict[str, set[str]] = field(
        default_factory=lambda: {
            'abbr': {'title'},
            'kbd': set(),
            'q': set(),
        }
    )
    """HTML elements that should be allowed in messages.

    The key is an element name,
    and the value is a set of attributes that are allowed on that element.
    All other elements and attributes will cause a test failure.
    (See also :py:attr:`~allowed_global_attributes`.)
    """

    allowed_global_attributes: set[str] = field(
        default_factory=lambda: {
            'dir',
            'lang',
        }
    )
    """HTML attributes that should be allowed on any element in messages.

    This is similar to :py:attr:`~allowed_html_elements`,
    but the given attribute names are allowed regardless of element name.
    """

    get_gender: Callable[[Any], Literal['m', 'f', 'n']] = get_gender_by_user_name
    """The ``get_gender`` function used by :py:class:`I18nFormatter`.

    Defaults to a function that expects the argument to be a user name,
    and looks it up on Meta-Wiki.
    You may specify a different function,
    where the type (the value you pass as user_name= into messages)
    can be anything you want.
    """

    check_translations: bool = True
    """Whether to check translations when they are loaded.

    By default, translations are checked as soon as they are loaded,
    and if there is a problem with the translations,
    an error is raised and the translations cannot be used.
    (This generally means that the tool cannot run;
    you will probably have to revert the latest localisation updates
    and fix the translation on translatewiki.net.)
    This protects against broken or even malicious messages.

    If you have set up Continuous Integration (CI),
    e.g. using GitLab CI or GitHub actions,
    and you are running ``pytest`` as part of your tests,
    then the translations checks will also be registered as tests
    (you should see various i18n/*.json files in pytest’s output).
    In this case, assuming CI also runs on translatewiki.net exports
    (and you won’t merge any localisation updates where CI fails),
    you can set this config to ``False`` to disable the runtime checks;
    this will speed up translation loading and therefore the tool’s startup
    (for a well-translated tool, by more than a second).

    Beware that, if you set this to ``False``,
    only the pytest integration in CI protects your tool from malicious translations.
    You must be confident that CI will run, and will run all pytest tests,
    and if possible you should configure your repository
    so that localisation updates cannot be merged if CI fails
    (in GitLab: Settings > Merge requests > Pipelines must succeed;
    no direct equivalent in GitHub).
    Otherwise, it is always safe to leave this set to ``True``.
    """


def mw2py(mw: str, locale: babel.Locale, variables: Sequence[str]) -> str:
    def replace_plural(match: re.Match[str]) -> str:
        nonlocal locale, variables
        number = int(match[1])
        variable = variables[number - 1]
        args = match[2].split('|')
        plurals = []
        tag_args = []
        for arg in args:
            key, _, text = arg.partition('=')
            if key.isnumeric():
                plurals.append(arg)
            else:
                tag_args.append(arg)
        tags = [tag for tag in ['zero', 'one', 'two', 'few', 'many'] if tag in locale.plural_form.tags]
        tags = tags[: len(tag_args) - 1] + ['other']
        for tag, tag_arg in zip(tags, tag_args, strict=False):
            plurals.append(f'{tag}={tag_arg}')
        return '{' + variable + '!p:' + ':'.join(plurals) + '}'

    py = re.sub(r'\{\{PLURAL:\$([1-9][0-9]*)\|([^}]*)\}\}', replace_plural, mw)

    def replace_gender(match: re.Match[str]) -> str:
        nonlocal variables
        number = int(match[1])
        variable = variables[number - 1]
        args = match[2].split('|')
        genders = []
        for gender, arg in zip(['m', 'f', 'n'], args, strict=False):
            genders.append(f'{gender}={arg}')
        return '{' + variable + '!g:' + ':'.join(genders) + '}'

    py = re.sub(r'\{\{GENDER:\$([1-9][0-9]*)\|([^}]*)\}\}', replace_gender, py)

    def replace_hyperlink(match: re.Match[str]) -> str:
        nonlocal variables
        number = int(match[1])
        variable = variables[number - 1]
        inner_html = match[2]
        assert '{' not in inner_html and '}' not in inner_html
        return '{' + variable + '!h:' + inner_html + '}'

    py = re.sub(r'\[\$([1-9][0-9]*) ([^]]*)\]', replace_hyperlink, py)

    def replace_unconverted(match: re.Match[str]) -> str:
        nonlocal variables
        number = int(match[1])
        variable = variables[number - 1]
        if variable.startswith('list_'):
            return '{' + variable + '!l}'
        else:
            return '{' + variable + '}'

    py = re.sub(r'\$([1-9][0-9]*)', replace_unconverted, py)

    return py


def load_translations(config: TranslationsConfig) -> tuple[dict[str, dict[str, str]], dict[str, str]]:
    translations: dict[str, dict[str, str]] = {}
    documentation: dict[str, str] = {}
    for entry in os.scandir(config.directory):
        if not entry.is_file():
            continue
        match = re.match(r'(.*)\.json$', entry.name)
        if not match:
            continue
        language = match[1]
        with open(entry.path, 'r') as f:
            data = json.load(f)
        if language == 'qqq':
            documentation = {key: value for key, value in data.items() if not key.startswith('@')}
            continue
        babel_language = config.language_code_to_babel(language)
        try:
            locale = babel.Locale(babel_language)
        except babel.UnknownLocaleError as e:
            if babel_language != language:
                note = f"The translation language code '{language}' was mapped to the Babel language code '{babel_language}', yet the "
            else:
                note = 'The '
            note += f"language code '{babel_language}' was not recognized by Babel. "
            if config.language_code_to_babel != language_code_to_babel:
                note += 'You may have to update your language_code_to_babel implementation in your tool_translations_config.'
            else:
                note += 'You may have to configure a custom language_code_to_babel implementation in your tool_translations_config.'
            e.add_note(note)
            raise
        translations[language] = {}
        for key in data:
            if key.startswith('@'):
                continue
            msg = mw2py(data[key], locale, config.variables.get(key, []))
            translations[language][key] = msg
        for key in config.derived_messages:
            source_key, transformation = config.derived_messages[key]
            if source_key in translations[language]:
                translations[language][key] = transformation(translations[language][source_key])
    if config.check_translations:
        from toolforge_i18n._translations_checks import check_all_translations

        check_all_translations(config, translations, documentation)
    return translations, documentation



import pytest

import toolforge_i18n._user_agent
from toolforge_i18n._language_info import lang_autonym, lang_dir, lang_fallbacks, lang_mw_to_bcp47

toolforge_i18n._user_agent.set_user_agent(  # noqa: SLF001
    'toolforge-i18n test (https://gitlab.wikimedia.org/lucaswerkmeister/toolforge_i18n/; mail@lucaswerkmeister.de)'
)


@pytest.mark.parametrize(
    'code, expected', [('en', 'English'), ('de', 'Deutsch'), ('fa', 'فارسی'), ('bn-x-Q6747180', None)]
)
def test_lang_autonym(code: str, expected: str | None) -> None:
    assert lang_autonym(code) == expected


@pytest.mark.parametrize('code, expected', [('en', 'en'), ('simple', 'en-simple'), ('unknown', 'unknown')])
def test_lang_mw_to_bcp47(code: str, expected: str) -> None:
    assert lang_mw_to_bcp47(code) == expected


@pytest.mark.parametrize('code, expected', [('en', 'ltr'), ('fa', 'rtl'), ('unknown', 'auto')])
def test_lang_dir(code: str, expected: str) -> None:
    assert lang_dir(code) == expected


@pytest.mark.parametrize(
    'code, expected',
    [
        ('en', []),
        ('de', []),
        ('de-at', ['de']),
        ('sh', ['sh-latn', 'sh-cyrl', 'bs', 'sr-el', 'sr-latn', 'hr']),
    ],
)
def test_lang_fallbacks(code: str, expected: list[str]) -> None:
    assert lang_fallbacks(code) == expected



from typing import Literal

import pytest

from toolforge_i18n._get_gender import get_gender_by_user_name


@pytest.mark.parametrize(
    'user, expected',
    [
        ('علاء', 'm'),
        ('Harmonia Amanda', 'f'),
        ('Nikki', 'n'),
        (None, 'n'),
    ],
)
def test_get_gender(user: str | None, expected: Literal['m', 'n', 'f']) -> None:
    assert get_gender_by_user_name(user) == expected
